//D and D Dice Roller
//Michael Stein - 10/22/2020

//Constants for serial communication with shift registers
const int dataPin=8;
const int latchPin=10;
const int clockPin=11;

//Constants for buttons
const int diceTypeBtn=5;
const int rollDiceBtn=6;

//Constants for Led indicators
const int diceTypeLed=4;  //This is the led that lights when we click the diceTypeBtn
const int rollLed=3;      //This is the led that lights when we click the rollDiceBtn

//dice type variables
boolean diceTypeDisplayed = false;
int button1State;             // the current reading from the input pin
int lastButton1State = HIGH;  // the previous reading from the input pin

//roll button vars
int button2State;             // the current reading from the input pin for the rollbtn
int lastButton2State = HIGH;  // the previous reading from the input pin for the rollbtn

// the following variables are unsigned longs because the time, measured in
// milliseconds, will quickly become a bigger number than can be stored in an int.
// These vars are used for debouncing the 2 buttons.
unsigned long lastDebounceTime1 = 0;  // the last time the output pin was toggled dice type button
unsigned long lastDebounceTime2 = 0;  // the last time the output pin was toggled for roll button
unsigned long debounceDelay = 50;    // the debounce time; increase if the output flickers

//Default to dice type to 20 (index 5) -the most frequently used
int currDiceIndex = 5;  //The dice types are 4,6,8,10,12,20 
int diceTypeArray[6] = {4,6,8,10,12,20};  

/* Shift registers display the individual LEDs in the 7 segment LED, 
 * We send binary data - 1 lights up an led, 0 does not.
These values work for common cathode leds. If you have common anode, reverse 1's and 0's.                      
"Truth table for common cathode 7 segment display"
The letters a - g in the LED sketch below correspond with the letters a-g in the Truth table below.
I did not wire up the decimal point.
    a
    --
f |    | b
    --
e |  g | c
    --
     d
Shift register pins:                 
 Qh g f e d c b a 
LED segments:      
   a b c d e f g dp      
#                          "0B" added    
0  1 1 1 1 1 1 0 0 11111100  0B11111100  
1  0 1 1 0 0 0 0 0 01100000  0B01100000  
2  1 1 0 1 1 0 1 0 11011010  0B11011010  
3  1 1 1 1 0 0 1 0 11110010  0B11110010  
4  0 1 1 0 0 1 1 0 01100110  0B01100110  
5  1 0 1 1 0 1 1 0 10110110  0B10110110  
6  1 0 1 1 1 1 1 0 10111110  0B10111110  
7  1 1 1 0 0 0 0 0 11100000  0B11100000  
8  1 1 1 1 1 1 1 0 11111110  0B11111110  
9  1 1 1 0 0 1 1 0 11100110  0B11100110  
0  1 1 1 1 1 1 0 0 11111100  0B11111100  
*/

// the binary numbers of 0-9 are initalized in an array
int numberArray[10]={0B11111100,0B01100000,0B11011010,0B11110010,
                  0B01100110,0B10110110,0B10111110,0B11100000,0B11111110,0B11100110};                  

int animationArray[16]={0B00000000,0B10000000,0B10000000,0B00000000,0B01000000,0B00000000
                      ,0B00100000,0B00000000,0B00010000,0B00000000,0B00000000,0B00010000
                      ,0B00000000,0B00001000,0B00000000,0B00000100};

//This struct is used to return the two digits from the function getTwoDigits below.
struct retValues {
  int firstNum;
  int secondNum;
};

//This value is used to "turn off" the first 7 segment display when displaying a single digit number.
int noDisplay = 0B00000000;

void setup() {
  // put your setup code here, to run once:
  pinMode(latchPin, OUTPUT);
  pinMode(dataPin, OUTPUT);
  pinMode(clockPin, OUTPUT);

  //setup buttons
  pinMode(diceTypeBtn,INPUT);
  pinMode(rollDiceBtn,INPUT);

  //setup leds
  pinMode(diceTypeLed, OUTPUT);
  pinMode(rollLed, OUTPUT);

  // if analog input pin 0 is unconnected, random analog
  // noise will cause the call to randomSeed() to generate
  // different seed numbers each time the sketch runs.
  // randomSeed() will then shuffle the random function.
  randomSeed(analogRead(0));
  
  //uncomment lines below for debugging with the serial monitor
  Serial.begin(9600);
  //struct retValues v = getTwoDigits(97);
  //Serial.println(v.firstNum);
  //Serial.println(v.secondNum);
  displayDiceType(); 
}

//getTwoDigits takes a number and breaks it down and returns two digits
struct retValues getTwoDigits(int iNum) {
  struct retValues r;

  if (iNum < 10) {
    r.firstNum = 0;
    r.secondNum = iNum;
    return r;
  }
  
  r.firstNum = iNum / 10;
  r.secondNum = iNum % 10;
  return r;  
}

//displayNumber takes a number and displays it on the two 7 seg displays
void displayNumber(int theNum) {
  digitalWrite(latchPin, LOW);
  struct retValues v = getTwoDigits(theNum);  
  shiftOut(dataPin, clockPin,MSBFIRST,numberArray[v.secondNum]);
  if (v.firstNum == 0 ) {
     shiftOut(dataPin, clockPin,MSBFIRST,noDisplay);
  } else {
     shiftOut(dataPin, clockPin,MSBFIRST,numberArray[v.firstNum]);
  }
        
  digitalWrite(latchPin, HIGH);
  return;
}

void displayDiceType () {
  int currentDiceType = diceTypeArray[currDiceIndex];
  
  //uncomment for debugging
  Serial.print("Dice type=");
  Serial.println(currentDiceType);
  displayNumber(currentDiceType);
  diceTypeDisplayed = true;
  digitalWrite(diceTypeLed, HIGH);
  digitalWrite(rollLed,LOW);  
}

void rollAnimation() {
  for (int a=0;a < 3; a++) {
    for (int x=0;x < 16; x=x+2) {
      digitalWrite(latchPin, LOW);
      shiftOut(dataPin, clockPin,MSBFIRST,animationArray[x]);
      shiftOut(dataPin, clockPin,MSBFIRST,animationArray[x+1]);
      digitalWrite(latchPin, HIGH);
      delay(25);
    }     
  }
}

void rollDice() {
  digitalWrite(diceTypeLed, LOW);
  digitalWrite(rollLed, HIGH);
  diceTypeDisplayed = false;

  //Determine which dice we are using
  int currentDiceType = diceTypeArray[currDiceIndex];
  int randNum;
  
  //These are the following types 4,6,8,10,12,20
  //Most dice have values from 1 through n (n being the dice type).
  //Exceptions are: D10 - numbered 0 - 9    
  switch (currentDiceType) {
    case 4:
      //random(min,max) returns a random number between min and max-1 (long)
      randNum = random(1,5);
      break;
    case 6:
      randNum = random(1,7);
      break;
    case 8:
      randNum = random(1,9);
      break;
    case 10:
      //notice D10 displays 0-9
      randNum = random(0,10);
      break;   
    case 12:
      randNum = random(1,13);
      break;     
    default:  //d20
      randNum = random(1,21);
      if (randNum == 20) {
        //critical hit!
        criticalHitSound();
      }
      break;
  }
  displayNumber(randNum);
  return;
}

void criticalHitSound() {
  //piezo buzzer
  int piezoPin = 12;
  for (int x = 0; x < 3;x++) {
    //"blink the display"    
    displayNumber(20);
    tone(piezoPin, 4000, 300);
    delay(300);

    digitalWrite(latchPin, LOW);
    shiftOut(dataPin, clockPin,MSBFIRST,noDisplay);
    shiftOut(dataPin, clockPin,MSBFIRST,noDisplay);
    digitalWrite(latchPin, HIGH);
    delay(300);      
  }
  
  return;     
}

void checkForDiceButtonClick() {
  if (digitalRead(diceTypeBtn) == HIGH) {
    if (diceTypeDisplayed) {      
      //The dice type is currently on display - bump it up to the next type
      if (currDiceIndex == 5) {
        currDiceIndex = 0;
      } else {
        currDiceIndex++;
      }
    }
    displayDiceType();
  }
}

void loop() {
  // read the state of the switch into a local variable:
  int reading1 = digitalRead(diceTypeBtn);
  int reading2 = digitalRead(rollDiceBtn);
  // check to see if someone pressed the button
  // (i.e. the input went from LOW to HIGH), and you've waited long enough
  // since the last press to ignore any noise:
 
  // If the switch changed, due to noise or pressing:
  if (reading1 != lastButton1State) {
    // reset the debouncing timer
    lastDebounceTime1 = millis();
  }

  // If the switch changed, due to noise or pressing:
  if (reading2 != lastButton2State) {
    // reset the debouncing timer
    lastDebounceTime2 = millis();
  }

  //For debouncing the dice type button.
  if ((millis() - lastDebounceTime1) > debounceDelay) {
    // whatever the reading is at, it's been there for longer than the debounce
    // delay, so take it as the actual current state:
 
    // if the button state has changed:
    if (reading1 != button1State) {
      button1State = reading1;
 
      // only call function if the new button state is HIGH
      if (button1State == HIGH) {
        checkForDiceButtonClick();  
      }
    }
  }

  //For debouncing the roll dice button.
  if ((millis() - lastDebounceTime2) > debounceDelay) {
    // whatever the reading is at, it's been there for longer than the debounce
    // delay, so take it as the actual current state:
 
    // if the button state has changed:
    if (reading2 != button2State) {
      button2State = reading2;
 
      // call function if the new button state is HIGH
      if (button2State == HIGH) {
        rollAnimation();
        rollDice();  
      }
    }
  }
  
  // save the reading. Next time through the loop, it'll be the lastButtonState:
  lastButton1State = reading1;
  lastButton2State = reading2;
  
}
